class CFG(object):

    def __init__(self):
        """ TODO """
        raise NotImplementedError

    def visualize(self):
        """ TODO """
        raise NotImplementedError

    def visualize_call_flow(self):
        """ TODO """
        raise NotImplementedError

    def visualize_instrs_per_funcs(self):
        """ TODO """
        raise NotImplementedError
